import functools
from typing import Callable, Coroutine

from gi.repository import Gio


def background_task(f: Callable[..., Coroutine]):
    """
    Wraps an async function to be run using ``Gio.Application.create_asyncio_task``.

    Useful to use async functions like signal handlers or GTK template callbacks.

    Note: The return value will be lost, so this is not suitable when you need to
    return something from the coroutine, what might be needed in some signal handlers.
    """

    @functools.wraps(f)
    def decor(*args, **kwargs):
        app = Gio.Application.get_default()
        app.create_asyncio_task(f(*args, **kwargs))

    return decor



class RequestError(Exception):
    """Exception raised when request fails."""


class ProviderError(Exception):
    """Exception raised when provider fails."""


class UnexpectedError(ProviderError):
    """Exception raised when provider fails."""


class APIKeyRequired(ProviderError):
    """Exception raised when provider fails."""


class APIKeyInvalid(ProviderError):
    """Exception raised when provider fails."""


class InvalidLangCode(ProviderError):
    """Exception raised when provider fails."""


class BatchSizeExceeded(ProviderError):
    """Exception raised when provider fails."""


class CharactersLimitExceeded(ProviderError):
    """Exception raised when provider fails."""


class ServiceLimitReached(ProviderError):
    """Exception raised when provider fails."""



# Copyright 2023 Mufeed Ali
# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
import concurrent.futures
from typing import Callable, TypeVar

from dialect.providers.base import BaseProvider


class LocalProvider(BaseProvider):
    """Base class for providers needing local threaded helpers"""

    _T = TypeVar("_T")

    async def run_async(self, worker: Callable[..., _T], *args) -> _T:
        """
        Runs worker in a ThreadPoolExecutor.

        Args:
            worker: Function to execute on the thread.
            *args: Args for the worker function.
        """

        loop = asyncio.get_running_loop()

        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(pool, worker, *args)



# Copyright 2020 gi-lom
# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# Copyright 2023 Libretto
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import re
from typing import Literal, TypedDict

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gst, Gtk, Spelling

from dialect.asyncio import background_task
from dialect.define import APP_ID, PROFILE, RES_PATH, TRANS_NUMBER
from dialect.languages import LanguagesListModel
from dialect.providers import (
    TRANSLATORS,
    TTS,
    APIKeyInvalid,
    APIKeyRequired,
    BaseProvider,
    ProviderError,
    RequestError,
    Translation,
    TranslationRequest,
)
from dialect.settings import Settings
from dialect.shortcuts import DialectShortcutsWindow
from dialect.utils import find_item_match, first_exclude
from dialect.widgets import LangSelector, SpeechButton, TextView, ThemeSwitcher


class _OngoingSpeech(TypedDict):
    text: str
    lang: str
    called_from: Literal["src", "dest"]


class _NotificationAction(TypedDict):
    label: str
    name: str


@Gtk.Template(resource_path=f"{RES_PATH}/window.ui")
class DialectWindow(Adw.ApplicationWindow):
    __gtype_name__ = "DialectWindow"

    # Properties
    translator_loading: bool = GObject.Property(type=bool, default=True)  # type: ignore
    selection_translation_queued: bool = GObject.Property(type=bool, default=False)  # type: ignore

    # Child widgets
    menu_btn: Gtk.MenuButton = Gtk.Template.Child()  # type: ignore
    main_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    error_page: Adw.StatusPage = Gtk.Template.Child()  # type: ignore
    translator_box: Gtk.Box = Gtk.Template.Child()  # type: ignore
    key_page: Adw.StatusPage = Gtk.Template.Child()  # type: ignore
    rmv_key_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore

    langs_button_box: Gtk.Box = Gtk.Template.Child()  # type: ignore
    switch_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    src_lang_selector: LangSelector = Gtk.Template.Child()  # type: ignore
    dest_lang_selector: LangSelector = Gtk.Template.Child()  # type: ignore

    return_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    forward_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore

    src_pron_revealer: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    src_pron_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    mistakes: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    mistakes_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    char_counter: Gtk.Label = Gtk.Template.Child()  # type: ignore
    src_text: TextView = Gtk.Template.Child()  # type: ignore
    clear_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    paste_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    src_speech_btn: SpeechButton = Gtk.Template.Child()  # type: ignore
    translate_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore

    dest_box: Gtk.Box = Gtk.Template.Child()  # type: ignore
    dest_pron_revealer: Gtk.Revealer = Gtk.Template.Child()  # type: ignore
    dest_pron_label: Gtk.Label = Gtk.Template.Child()  # type: ignore
    dest_text: TextView = Gtk.Template.Child()  # type: ignore
    dest_toolbar_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    trans_spinner: Adw.Spinner = Gtk.Template.Child()  # type: ignore
    trans_warning: Gtk.Image = Gtk.Template.Child()  # type: ignore
    edit_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    copy_btn: Gtk.Button = Gtk.Template.Child()  # type: ignore
    dest_speech_btn: SpeechButton = Gtk.Template.Child()  # type: ignore

    toast: Adw.Toast | None = None  # for notification management
    toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()  # type: ignore

    win_key_ctrlr: Gtk.EventControllerKey = Gtk.Template.Child()  # type: ignore

    # Providers objects
    provider: dict[str, BaseProvider | None] = {"trans": None, "tts": None}

    # Text to speech
    speech_provider_failed = False  # tts provider loading failed
    current_speech: _OngoingSpeech | None = None
    speech_loading = False  # tts loading status

    # Preset language values
    src_langs: list[str] = []
    dest_langs: list[str] = []

    current_history = 0  # for history management

    # Translation-related variables
    selection_translation_langs: tuple[str | None, str | None] = (None, None)
    next_translation: TranslationRequest | None = None  # for ongoing translation
    translation_loading = False  # for ongoing translation

    # Suggestions
    before_suggest: str | None = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Application object
        self.app: Adw.Application = kwargs["application"]

        # GStreamer playbin object and related setup
        self.player = Gst.ElementFactory.make("playbin", "player")
        if self.player:
            if bus := self.player.get_bus():
                bus.add_signal_watch()
                bus.connect("message", self._on_gst_message)

        # Setup window
        self.setup_actions()
        self.setup()

    def setup_actions(self):
        back = Gio.SimpleAction(name="back")
        back.props.enabled = False
        back.connect("activate", self._on_back_action)
        self.add_action(back)

        forward_action = Gio.SimpleAction(name="forward")
        forward_action.props.enabled = False
        forward_action.connect("activate", self._on_forward_action)
        self.add_action(forward_action)

        switch_action = Gio.SimpleAction(name="switch")
        switch_action.connect("activate", self._on_switch_action)
        self.add_action(switch_action)

        from_action = Gio.SimpleAction(name="from")
        from_action.connect("activate", self._on_from_action)
        self.add_action(from_action)

        to_action = Gio.SimpleAction(name="to")
        to_action.connect("activate", self._on_to_action)
        self.add_action(to_action)

        clear_action = Gio.SimpleAction(name="clear")
        clear_action.props.enabled = False
        clear_action.connect("activate", self._on_clear_action)
        self.add_action(clear_action)

        font_size_inc_action = Gio.SimpleAction(name="font-size-inc")
        font_size_inc_action.connect("activate", self._on_font_size_inc_action)
        self.add_action(font_size_inc_action)

        font_size_dec_action = Gio.SimpleAction(name="font-size-dec")
        font_size_dec_action.connect("activate", self._on_font_size_dec_action)
        self.add_action(font_size_dec_action)

        paste_action = Gio.SimpleAction(name="paste")
        paste_action.connect("activate", self._on_paste_action)
        self.add_action(paste_action)

        copy_action = Gio.SimpleAction(name="copy")
        copy_action.props.enabled = False
        copy_action.connect("activate", self._on_copy_action)
        self.add_action(copy_action)

        listen_dest_action = Gio.SimpleAction(name="listen-dest")
        listen_dest_action.connect("activate", self._on_dest_listen_action)
        listen_dest_action.props.enabled = False
        self.add_action(listen_dest_action)

        suggest_action = Gio.SimpleAction(name="suggest")
        suggest_action.props.enabled = False
        suggest_action.connect("activate", self._on_suggest_action)
        self.add_action(suggest_action)

        suggest_ok_action = Gio.SimpleAction(name="suggest-ok")
        suggest_ok_action.connect("activate", self._on_suggest_ok_action)
        self.add_action(suggest_ok_action)

        suggest_cancel_action = Gio.SimpleAction(name="suggest-cancel")
        suggest_cancel_action.connect("activate", self._on_suggest_cancel_action)
        self.add_action(suggest_cancel_action)

        listen_src_action = Gio.SimpleAction(name="listen-src")
        listen_src_action.connect("activate", self._on_src_listen_action)
        listen_src_action.props.enabled = False
        self.add_action(listen_src_action)

        translation_action = Gio.SimpleAction(name="translation")
        translation_action.props.enabled = False
        translation_action.connect("activate", self._on_translation)
        self.add_action(translation_action)

    def setup(self):
        self.set_default_icon_name(APP_ID)

        # Set devel style
        if PROFILE == "Devel":
            self.add_css_class("devel")

        # Theme Switcher
        theme_switcher = ThemeSwitcher()
        menu: Gtk.PopoverMenu = self.menu_btn.props.popover  # type: ignore
        menu.add_child(theme_switcher, "theme")

        # Save settings on close
        self.connect("unrealize", self.save_settings)

        self.setup_selectors()
        self.setup_translation()
        self.setup_spell_checking()
        self.set_help_overlay(DialectShortcutsWindow())

        # Load translator
        self.load_translator()
        # Load text to speech
        self.load_tts()

        # Listen to active providers changes
        Settings.get().connect("provider-changed", self._on_active_provider_changed)

        # Bind text views font size
        self.src_text.bind_property("font-size", self.dest_text, "font-size", GObject.BindingFlags.BIDIRECTIONAL)

        # Set initial saved text view font size
        if Settings.get().custom_default_font_size:
            font_size = Settings.get().default_font_size
            self.set_font_size(font_size)

        # Set src textview mod key requirement
        self.src_text.activate_mod = not bool(Settings.get().translate_accel_value)
        Settings.get().connect(
            "changed::translate-accel",
            lambda s, _k: self.src_text.set_property("activate_mod", not bool(s.translate_accel_value)),
        )

    def setup_spell_checking(self):
        # Enable spell-checking
        self.spell_checker: Spelling.Checker = Spelling.Checker.get_default()
        spell_checker_adapter = Spelling.TextBufferAdapter.new(self.src_buffer, self.spell_checker)
        spell_checker_menu = spell_checker_adapter.get_menu_model()
        self.src_text.set_extra_menu(spell_checker_menu)
        self.src_text.insert_action_group("spelling", spell_checker_adapter)
        spell_checker_adapter.set_enabled(True)

        # Collect the spell checking provider's supported languages.
        self.spell_checker_supported_languages = {}
        for lang_object in self.spell_checker.get_provider().list_languages():
            lang_code = lang_object.get_code()
            lang_base_code = re.split("_|-", lang_code)[0]
            if lang_base_code not in self.spell_checker_supported_languages:
                self.spell_checker_supported_languages[lang_base_code] = []
            self.spell_checker_supported_languages[lang_base_code].append(lang_code)

    def setup_selectors(self):
        def lang_names_func(code: str):
            return self.provider["trans"].get_lang_name(code)  # type: ignore

        # Languages models
        self.src_lang_model = LanguagesListModel(lang_names_func)
        self.src_recent_lang_model = LanguagesListModel(lang_names_func)
        self.dest_lang_model = LanguagesListModel(lang_names_func)
        self.dest_recent_lang_model = LanguagesListModel(lang_names_func)

        # Src lang selector
        self.src_lang_selector.bind_models(self.src_lang_model, self.src_recent_lang_model)
        # Dest lang selector
        self.dest_lang_selector.bind_models(self.dest_lang_model, self.dest_recent_lang_model)

    def setup_translation(self):
        # Src buffer
        self.src_buffer = self.src_text.props.buffer
        self.src_buffer.connect("changed", self._on_src_text_changed)
        self.src_buffer.connect("end-user-action", self._on_user_action_ended)

        # Dest buffer
        self.dest_buffer = self.dest_text.props.buffer
        self.dest_buffer.props.text = ""
        self.dest_buffer.connect("changed", self._on_dest_text_changed)

        # Translation progress
        self.trans_spinner.hide()
        self.trans_warning.hide()

    def reload_provider(self, kind: str):
        match kind:
            case "translator":
                self.load_translator()
            case "tts":
                self.load_tts()

    @background_task
    async def load_translator(self):
        self.translator_loading = True

        provider = Settings.get().active_translator

        # Show loading view
        self.main_stack.props.visible_child_name = "loading"

        # Translator object
        self.provider["trans"] = TRANSLATORS[provider]()
        # Get saved languages
        self.src_langs = self.provider["trans"].recent_src_langs
        self.dest_langs = self.provider["trans"].recent_dest_langs
        # Connect to provider settings changes
        self.provider["trans"].settings.connect(
            "changed::instance-url", self._on_provider_changed, self.provider["trans"].name
        )
        self.provider["trans"].settings.connect(
            "changed::api-key", self._on_provider_changed, self.provider["trans"].name
        )

        try:
            # Do provider init
            await self.provider["trans"].init_trans()

            # Update navigation UI
            self._check_navigation_enabled()
            # Check mistakes support
            self._check_mistakes()
            # Check pronunciation support
            self._check_pronunciation()
            # Check suggestions support and update UI
            self._on_suggest_cancel_action()
            self.edit_btn.props.visible = self.provider["trans"].supports_suggestions

            # Update langs
            self.src_lang_model.set_langs(self.provider["trans"].src_languages)
            self.dest_lang_model.set_langs(self.provider["trans"].dest_languages)

            # Update selected langs
            set_auto = Settings.get().src_auto and self.provider["trans"].supports_detection
            src_lang = self.provider["trans"].src_languages[0]
            if self.src_langs and self.src_langs[0] in self.provider["trans"].src_languages:
                src_lang = self.src_langs[0]
            self.src_lang_selector.selected = "auto" if set_auto else src_lang

            dest_lang = self.provider["trans"].dest_languages[1]
            if self.dest_langs and self.dest_langs[0] in self.provider["trans"].dest_languages:
                dest_lang = self.dest_langs[0]
            self.dest_lang_selector.selected = dest_lang

            # Update chars limit
            if self.provider["trans"].chars_limit == -1:  # -1 means unlimited
                self.char_counter.props.label = ""
            else:
                count = f"{str(self.src_buffer.get_char_count())}/{self.provider['trans'].chars_limit}"
                self.char_counter.props.label = count

            # Check API key
            if self.provider["trans"].supports_api_key:
                if self.provider["trans"].api_key:
                    try:
                        if await self.provider["trans"].validate_api_key(self.provider["trans"].api_key):
                            self.main_stack.props.visible_child_name = "translate"
                        else:
                            self.show_translator_api_key_view()
                    except ProviderError or RequestError as exc:
                        logging.error(exc, exc_info=exc)
                        self.show_translator_error_view(detail=str(exc))
                elif not self.provider["trans"].api_key and self.provider["trans"].api_key_required:
                    self.show_translator_api_key_view(required=True)
                else:
                    self.main_stack.props.visible_child_name = "translate"
            else:
                self.main_stack.props.visible_child_name = "translate"

        # Loading failed
        except (RequestError, ProviderError) as exc:
            logging.error(exc)

            # API key error
            if isinstance(exc, APIKeyRequired):
                self.show_translator_api_key_view(required=True)
            elif isinstance(exc, APIKeyInvalid):
                self.show_translator_api_key_view()

            # Other errors
            else:
                service = self.provider["trans"].prettyname
                url = self.provider["trans"].instance_url
                detail = str(exc)

                if isinstance(exc, RequestError):
                    title = _("Couldn’t connect to the translation service")
                    description = _("We can’t connect to the server. Please check for network issues.")
                    if self.provider["trans"].supports_instances:
                        description = _(
                            (
                                "We can’t connect to the {service} instance “{url}“.\n"
                                "Please check for network issues or if the address is correct."
                            )
                        ).format(service=service, url=url)
                    self.show_translator_error_view(title, description, detail)
                elif self.provider["trans"].supports_instances:
                    description = _(
                        (
                            "Failed loading “{url}“, check if the instance address is correct or report in the Dialect bug tracker"
                            " if the issue persists."
                        )
                    ).format(url=url)
                    self.show_translator_error_view(description=description, detail=detail)
                else:
                    self.show_translator_error_view(detail=detail)

        finally:
            self.translator_loading = False

    def show_translator_error_view(
        self,
        title: str = _("Failed loading the translation service"),
        description: str = _("Please report this in the Dialect bug tracker if the issue persists."),
        detail: str | None = None,
    ):
        if detail:  # Add detail bellow description
            if description:
                description += "\n\n"
            description += "<small><tt>" + detail + "</tt></small>"

        self.error_page.props.title = title
        self.error_page.props.description = description

        self.main_stack.props.visible_child_name = "error"

    def show_translator_api_key_view(self, required=False):
        if not self.provider["trans"]:
            return

        if required:
            self.key_page.props.title = _("API key is required to use the service")
            self.key_page.props.description = _("Please set an API key in the preferences.")

        else:
            self.key_page.props.title = _("The provided API key is invalid")
            if self.provider["trans"].api_key_required:
                self.key_page.props.description = _("Please set a valid API key in the preferences.")
            else:
                self.key_page.props.description = _(
                    "Please set a valid API key or unset the API key in the preferences."
                )
                self.rmv_key_btn.props.visible = True

        self.main_stack.props.visible_child_name = "api-key"

    @background_task
    async def load_tts(self):
        self.src_speech_btn.loading()
        self.dest_speech_btn.loading()

        # TTS name
        provider = Settings.get().active_tts

        # Check if TTS is disabled
        if provider != "":
            self.src_speech_btn.props.visible = True
            self.dest_speech_btn.props.visible = True

            # TTS Object
            self.provider["tts"] = TTS[provider]()
            # Connect to provider settings changes
            self.provider["tts"].settings.connect(
                "changed::instance-url", self._on_provider_changed, self.provider["tts"].name
            )
            self.provider["tts"].settings.connect(
                "changed::api-key", self._on_provider_changed, self.provider["tts"].name
            )

            try:
                # Do TTS init
                await self.provider["tts"].init_tts()

                self.speech_provider_failed = False
                self.src_speech_btn.ready()
                self.dest_speech_btn.ready()
                self._check_speech_enabled()

            # Loading failed
            except (RequestError, ProviderError) as exc:
                logging.error(exc)

                button_text = _("Failed loading the text-to-speech service. Retry?")
                toast_text = _("Failed loading the text-to-speech service")
                if isinstance(exc, RequestError):
                    toast_text = _("Failed loading the text-to-speech service, check for network issues")

                self.speech_provider_failed = True
                self.src_speech_btn.error(button_text)
                self.dest_speech_btn.error(button_text)
                self.send_notification(toast_text)
                self._check_speech_enabled()

        else:
            self.provider["tts"] = None
            self.src_speech_btn.props.visible = False
            self.dest_speech_btn.props.visible = False

    def translate(self, text: str, src_lang: str | None, dest_lang: str | None):
        """
        Translates the given text from auto detected language to last used
        language
        """
        if not self.provider["trans"]:
            return

        # Set src lang to Auto
        if src_lang is None:
            self.src_lang_selector.selected = "auto"
        else:
            self.src_lang_selector.selected = src_lang
        if dest_lang is not None and dest_lang in self.provider["trans"].dest_languages:
            self.dest_lang_selector.selected = dest_lang
            self.dest_lang_selector.emit("user-selection-changed")
        # Set text to src buffer
        self.src_buffer.props.text = text
        # Run translation
        self._on_translation()

    @background_task
    async def translate_selection(self, src_lang: str | None, dest_lang: str | None):
        """Runs `translate` with the selection clipboard text"""
        if display := Gdk.Display.get_default():
            clipboard = display.get_primary_clipboard()
            try:
                if text := await clipboard.read_text_async():  # type: ignore
                    self.translate(text, src_lang, dest_lang)
            except GLib.Error as exc:
                logging.error(exc)
                self.send_notification(_("Couldn’t read selection clip board!"))

    def queue_selection_translation(self, src_lang: str | None, dest_lang: str | None):
        """Call `translate_selection` or queue it until the window is focused"""
        if self.props.is_active:
            self.translate_selection(src_lang, dest_lang)
        else:
            self.selection_translation_queued = True
            self.selection_translation_langs = (src_lang, dest_lang)

    def save_settings(self, *args, **kwargs):
        if not self.is_maximized():
            size = self.get_default_size()
            Settings.get().window_size = (size.width, size.height)  # type: ignore
        if self.provider["trans"] is not None:
            self.provider["trans"].recent_src_langs = self.src_langs
            self.provider["trans"].recent_dest_langs = self.dest_langs

    def send_notification(
        self,
        text: str,
        queue: bool | None = False,
        action: _NotificationAction | None = None,
        timeout=5,
        priority=Adw.ToastPriority.NORMAL,
    ):
        """
        Display an in-app notification.

        Args:
            text: The text or message of the notification.
            queue: If True, the notification will be queued.
            action: A dict containing the action to be called.
            timeout: Toast timeout.
            timeout: Toast priority.
        """

        def toast_dismissed(_toast: Adw.Toast):
            self.toast = None

        if not queue and self.toast is not None:
            self.toast.dismiss()
        self.toast = Adw.Toast(title=text)
        self.toast.connect("dismissed", toast_dismissed)
        if action is not None:
            self.toast.props.button_label = action["label"]
            self.toast.props.action_name = action["name"]
        self.toast.props.timeout = timeout
        self.toast.props.priority = priority
        self.toast_overlay.add_toast(self.toast)

    def set_font_size(self, size: int):
        self.src_text.font_size = size

    def add_history_entry(self, translation: Translation):
        """Add a history entry to the history list."""
        if not self.provider["trans"]:
            return

        if self.current_history > 0:
            del self.provider["trans"].history[: self.current_history]
            self.current_history = 0
        if len(self.provider["trans"].history) == TRANS_NUMBER:
            self.provider["trans"].history.pop()
        self.provider["trans"].history.insert(0, translation)
        self._check_navigation_enabled()

    @property
    def current_translation(self) -> Translation | None:
        """Get the current active translation, respecting the history navigation"""
        if not self.provider["trans"]:
            return None

        try:
            return self.provider["trans"].history[self.current_history]
        except IndexError:
            return None

    def _check_navigation_enabled(self):
        self.lookup_action("back").props.enabled = self.current_history < len(self.provider["trans"].history) - 1  # type: ignore
        self.lookup_action("forward").props.enabled = self.current_history > 0  # type: ignore

    def _check_mistakes(self):
        if not self.provider["trans"]:
            return

        translation = self.current_translation
        if self.provider["trans"].supports_mistakes and translation and translation.mistakes:
            self.mistakes_label.set_markup(_("Did you mean: ") + f'<a href="#">{translation.mistakes.markup}</a>')
            self.mistakes.props.reveal_child = True
        elif self.mistakes.props.reveal_child:
            self.mistakes.props.reveal_child = False

    def _check_pronunciation(self):
        if not self.provider["trans"]:
            return

        if not self.provider["trans"].supports_pronunciation:
            self.src_pron_revealer.props.reveal_child = False
            self.dest_pron_revealer.props.reveal_child = False
            self.app.lookup_action("pronunciation").props.enabled = False  # type: ignore
        else:
            self.app.lookup_action("pronunciation").props.enabled = True  # type: ignore
            reveal = Settings.get().show_pronunciation
            translation = self.current_translation

            if translation and translation.pronunciation.src and not translation.mistakes:
                self.src_pron_label.props.label = translation.pronunciation.src
                self.src_pron_revealer.props.reveal_child = reveal
            elif self.src_pron_revealer.props.reveal_child:
                self.src_pron_revealer.props.reveal_child = False

            if translation and translation.pronunciation.dest:
                self.dest_pron_label.props.label = translation.pronunciation.dest
                self.dest_pron_revealer.props.reveal_child = reveal
            elif self.dest_pron_revealer.props.reveal_child:
                self.dest_pron_revealer.props.reveal_child = False

    def _check_speech_enabled(self):
        if not self.provider["tts"]:
            return

        src_playing = dest_playing = False
        if self.current_speech:
            src_playing = self.current_speech["called_from"] == "src"
            dest_playing = self.current_speech["called_from"] == "dest"

        # Check src listen button
        self.lookup_action("listen-src").set_enabled(  # type: ignore
            self.speech_provider_failed
            or self.src_lang_selector.selected in self.provider["tts"].tts_languages
            and self.src_buffer.get_char_count() != 0
            and not self.speech_loading
            and not dest_playing
        )

        # Check dest listen button
        self.lookup_action("listen-dest").set_enabled(  # type: ignore
            self.speech_provider_failed
            or self.dest_lang_selector.selected in self.provider["tts"].tts_languages
            and self.dest_buffer.get_char_count() != 0
            and not self.speech_loading
            and not src_playing
        )

    def _check_switch_enabled(self):
        if not self.provider["trans"]:
            return

        # Disable or enable switch function.
        self.lookup_action("switch").props.enabled = (  # type: ignore
            self.src_lang_selector.selected in self.provider["trans"].dest_languages
            and self.dest_lang_selector.selected in self.provider["trans"].src_languages
        )

    def _pick_spell_checking_language(self, lang_code: str):
        # Try and set the correct language if available.
        lang_base_code = lang_code.split("-")[0]
        default_user_lang_code = self.spell_checker.get_provider().get_default_code()

        spell_checker_lang_code = default_user_lang_code  # Default to the user's preference

        if lang_base_code in self.spell_checker_supported_languages:
            if lang_code in self.spell_checker_supported_languages[lang_base_code]:
                # The language is matched exactly.
                spell_checker_lang_code = lang_code
            elif (code := lang_code.replace("-", "_")) in self.spell_checker_supported_languages[lang_base_code]:
                # The language code needs underscores within the provider.
                spell_checker_lang_code = code
            elif default_user_lang_code.startswith(lang_base_code):
                # Default to user preference if at least the base language code matches.
                # Probably the most common scenario: en -> en_US.
                spell_checker_lang_code = default_user_lang_code
            else:
                # Try to set the language even if the country code doesn't match, to the first available one.
                spell_checker_lang_code = self.spell_checker_supported_languages[lang_base_code][0]

        self.spell_checker.set_language(spell_checker_lang_code)

    """
    User interface functions
    """

    def _on_back_action(self, *_args):
        """Go back one step in history."""
        if self.current_history != TRANS_NUMBER:
            self.current_history += 1
            self._history_update()

    def _on_forward_action(self, *_args):
        """Go forward one step in history."""
        if self.current_history != 0:
            self.current_history -= 1
            self._history_update()

    def _history_update(self):
        if not self.provider["trans"]:
            return

        if translation := self.current_translation:
            self.src_lang_selector.selected = translation.original.src
            self.dest_lang_selector.selected = translation.original.dest
            self.src_buffer.props.text = translation.original.text
            self.dest_buffer.props.text = translation.text

            self._check_navigation_enabled()
            self._check_mistakes()
            self._check_pronunciation()

    def _on_switch_action(self, *_args):
        # Get variables
        self.langs_button_box.props.sensitive = False
        self.lookup_action("translation").props.enabled = False  # type: ignore
        src_language = self.src_lang_selector.selected
        dest_language = self.dest_lang_selector.selected
        src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)
        if src_language == "auto":
            return

        # Switch all
        self.src_lang_selector.selected = dest_language
        self.dest_lang_selector.selected = src_language
        self.src_buffer.props.text = dest_text
        self.dest_buffer.props.text = src_text
        self.add_history_entry(Translation(src_text, TranslationRequest(dest_text, src_language, dest_language)))

        # Re-enable widgets
        self.langs_button_box.props.sensitive = True
        self.lookup_action("translation").props.enabled = self.src_buffer.get_char_count() != 0  # type: ignore

    def _on_from_action(self, *_args):
        self.src_lang_selector.button.popup()

    def _on_to_action(self, *_args):
        self.dest_lang_selector.button.popup()

    def _on_clear_action(self, *_args):
        self.src_buffer.props.text = ""
        self.src_buffer.emit("end-user-action")

    def _on_font_size_inc_action(self, *_args):
        self.src_text.font_size_inc()

    def _on_font_size_dec_action(self, *_args):
        self.src_text.font_size_dec()

    def _on_copy_action(self, *_args):
        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)
        if display := Gdk.Display.get_default():
            display.get_clipboard().set(dest_text)
            self.send_notification(_("Copied to clipboard"), timeout=1)

    @background_task
    async def _on_paste_action(self, *_args):
        if display := Gdk.Display.get_default():
            clipboard = display.get_clipboard()
            if text := await clipboard.read_text_async():  # type: ignore
                end_iter = self.src_buffer.get_end_iter()
                self.src_buffer.insert(end_iter, text)
                self.src_buffer.emit("end-user-action")

    def _on_suggest_action(self, *_args):
        self.dest_toolbar_stack.props.visible_child_name = "edit"
        self.before_suggest = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True
        )
        self.dest_text.props.editable = True

    @background_task
    async def _on_suggest_ok_action(self, *_args):
        if not self.provider["trans"]:
            return

        try:
            dest_text = self.dest_buffer.get_text(
                self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True
            )
            if translation := self.current_translation:
                if await self.provider["trans"].suggest(
                    translation.original.text, translation.original.src, translation.original.dest, dest_text
                ):
                    self.send_notification(_("New translation has been suggested!"))
                else:
                    self.send_notification(_("Suggestion failed."))

        except (RequestError, ProviderError) as exc:
            logging.error(exc)
            self.send_notification(_("Suggestion failed."))

        finally:
            self.dest_toolbar_stack.props.visible_child_name = "default"
            self.dest_text.props.editable = False
            self.before_suggest = None

    def _on_suggest_cancel_action(self, *_args):
        self.dest_toolbar_stack.props.visible_child_name = "default"
        if self.before_suggest is not None:
            self.dest_buffer.props.text = self.before_suggest
            self.before_suggest = None
        self.dest_text.props.editable = False

    def _on_src_listen_action(self, *_args):
        if self.current_speech:
            self._speech_reset()
            return

        src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
        src_language = self.src_lang_selector.selected
        self._on_speech(src_text, src_language, "src")

    def _on_dest_listen_action(self, *_args):
        if self.current_speech:
            self._speech_reset()
            return

        dest_text = self.dest_buffer.get_text(self.dest_buffer.get_start_iter(), self.dest_buffer.get_end_iter(), True)
        dest_language = self.dest_lang_selector.selected
        self._on_speech(dest_text, dest_language, "dest")

    @background_task
    async def _on_speech(self, text: str, lang: str, called_from: Literal["src", "dest"]):
        # Retry loading TTS provider
        if self.speech_provider_failed:
            self.load_tts()
            return

        if not text or not self.provider["tts"] or not self.player:
            return

        # Set loading state and current speech to update UI
        self.speech_loading = True
        self.current_speech = {"text": text, "lang": lang, "called_from": called_from}
        self._check_speech_enabled()

        if called_from == "src":  # Show spinner on button
            self.src_speech_btn.loading()
        else:
            self.dest_speech_btn.loading()

        # Download speech
        try:
            file_ = await self.provider["tts"].speech(self.current_speech["text"], self.current_speech["lang"])
            uri = "file://" + file_.name
            self.player.set_property("uri", uri)
            self.player.set_state(Gst.State.PLAYING)
            self.add_tick_callback(self._gst_progress_timeout)
            file_.close()

        except (RequestError, ProviderError) as exc:
            logging.error(exc)

            text = _("Text-to-Speech failed")
            action: _NotificationAction | None = None

            if isinstance(exc, RequestError):
                text = _("Text-to-Speech failed, check for network issues")

            if self.current_speech:
                called_from = self.current_speech["called_from"]
                action = {
                    "label": _("Retry"),
                    "name": "win.listen-src" if called_from == "src" else "win.listen-dest",
                }

                button_text = _("Text-to-Speech failed. Retry?")
                if called_from == "src":
                    self.src_speech_btn.error(button_text)
                else:
                    self.dest_speech_btn.error(button_text)

            self.send_notification(text, action=action)
            self._speech_reset(False)

    def _speech_reset(self, set_ready: bool = True):
        if not self.player:
            return

        self.player.set_state(Gst.State.NULL)
        self.current_speech = None
        self.speech_loading = False
        self._check_speech_enabled()

        if set_ready:
            self.src_speech_btn.ready()
            self.dest_speech_btn.ready()

    def _on_gst_message(self, _bus, message: Gst.Message):
        if message.type == Gst.MessageType.EOS or message.type == Gst.MessageType.ERROR:
            if message.type == Gst.MessageType.ERROR:
                logging.error("Some error occurred while trying to play.")
            self._speech_reset()

    def _gst_progress_timeout(self, _widget, _clock):
        if not self.player:
            return False

        if self.current_speech and self.player.get_state(Gst.CLOCK_TIME_NONE) != Gst.State.NULL:
            have_pos, pos = self.player.query_position(Gst.Format.TIME)
            have_dur, dur = self.player.query_duration(Gst.Format.TIME)

            if have_pos and have_dur:
                if self.current_speech["called_from"] == "src":
                    self.src_speech_btn.progress(pos / dur)
                else:
                    self.dest_speech_btn.progress(pos / dur)

                if self.speech_loading:
                    self.speech_loading = False
                    self._check_speech_enabled()

            return True

        return False

    def _on_src_text_changed(self, buffer: Gtk.TextBuffer):
        if not self.provider["trans"]:
            return

        char_count = buffer.get_char_count()

        # If the text is over the highest number of characters allowed, it is truncated.
        # This is done for avoiding exceeding the limit imposed by translation services.
        if self.provider["trans"].chars_limit == -1:  # -1 means unlimited
            self.char_counter.props.label = ""
        else:
            self.char_counter.props.label = f"{str(char_count)}/{self.provider['trans'].chars_limit}"

            if char_count >= self.provider["trans"].chars_limit:
                self.send_notification(_("{} characters limit reached!").format(self.provider["trans"].chars_limit))
                buffer.delete(buffer.get_iter_at_offset(self.provider["trans"].chars_limit), buffer.get_end_iter())

        sensitive = char_count != 0
        self.lookup_action("translation").props.enabled = sensitive  # type: ignore
        self.lookup_action("clear").props.enabled = sensitive  # type: ignore
        self._check_speech_enabled()

    def _on_dest_text_changed(self, buffer: Gtk.TextBuffer):
        if not self.provider["trans"]:
            return

        sensitive = buffer.get_char_count() != 0
        self.lookup_action("copy").props.enabled = sensitive  # type: ignore
        self.lookup_action("suggest").set_enabled(  # type: ignore
            self.provider["trans"].supports_suggestions and sensitive
        )
        self._check_speech_enabled()

    def _on_user_action_ended(self, _buffer):
        if Settings.get().live_translation:
            self._on_translation()

    @Gtk.Template.Callback()
    def _on_is_active_changed(self, *_args):
        if self.selection_translation_queued and self.props.is_active:
            src, dest = self.selection_translation_langs
            self.selection_translation_queued = False
            self.selection_translation_langs = (None, None)
            self.translate_selection(src, dest)

    @Gtk.Template.Callback()
    def _on_retry_load_translator_clicked(self, *_args):
        self.reload_provider("translator")

    @Gtk.Template.Callback()
    def _on_remove_key_and_reload_clicked(self, *_args):
        if self.provider["trans"]:
            self.provider["trans"].reset_api_key()
        self.reload_provider("translator")

    @Gtk.Template.Callback()
    def _on_src_lang_changed(self, *_args):
        """Called on self.src_lang_selector::notify::selected signal"""
        if not self.provider["trans"]:
            return

        code = self.src_lang_selector.selected
        dest_code = self.dest_lang_selector.selected

        if self.provider["trans"].cmp_langs(code, dest_code):
            # Get first lang from saved src langs that is not current dest
            if valid := first_exclude(self.src_langs, dest_code):
                # Check if it's a valid dest lang
                valid = find_item_match([valid], self.provider["trans"].dest_languages)
            if not valid:  # If not, just get the first lang from the list that is not selected
                valid = first_exclude(self.provider["trans"].dest_languages, dest_code)

            self.dest_lang_selector.selected = valid or ""

        if code in self.provider["trans"].src_languages:
            # Update saved src langs list
            if code in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code)
            elif code.lower() in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code.lower())
            elif len(self.src_langs) == 4:
                self.src_langs.pop()
            self.src_langs.insert(0, code)

        # Try to set the language in the spell checker.
        self._pick_spell_checking_language(code)

        # Rewrite recent langs
        self.src_recent_lang_model.set_langs(self.src_langs, auto=True)

        self._check_switch_enabled()
        self._check_speech_enabled()

    @Gtk.Template.Callback()
    def _on_dest_lang_changed(self, *_args):
        """Called on self.dest_lang_selector::notify::selected signal"""
        if not self.provider["trans"]:
            return

        code = self.dest_lang_selector.selected
        src_code = self.src_lang_selector.selected

        if self.provider["trans"].cmp_langs(code, src_code):
            # Get first lang from saved dest langs that is not current src
            if valid := first_exclude(self.dest_langs, src_code):
                # Check if it's a valid src lang
                valid = find_item_match([valid], self.provider["trans"].src_languages)
            if not valid:  # If not, just get the first lang from the list that is not selected
                valid = first_exclude(self.provider["trans"].src_languages, src_code)

            self.src_lang_selector.selected = valid or ""

        # Update saved dest langs list
        if code in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code)
        elif code.lower() in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code.lower())
        elif len(self.src_langs) == 4:
            self.dest_langs.pop()
        self.dest_langs.insert(0, code)

        # Rewrite recent langs
        self.dest_recent_lang_model.set_langs(self.dest_langs)

        self._check_switch_enabled()
        self._check_speech_enabled()

    @Gtk.Template.Callback()
    def _on_key_event(self, _ctrl, keyval: int, _keycode: int, state: Gdk.ModifierType):
        """Called on self.win_key_ctrlr::key-pressed signal"""
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        shift_mask = Gdk.ModifierType.SHIFT_MASK
        unicode_key_val = Gdk.keyval_to_unicode(keyval)
        if (
            GLib.unichar_isgraph(chr(unicode_key_val))
            and modifiers in (shift_mask, 0)
            and not self.dest_text.props.editable
            and not self.src_text.is_focus()
        ):
            self.src_text.grab_focus()
            end_iter = self.src_buffer.get_end_iter()
            self.src_buffer.insert(end_iter, chr(unicode_key_val))
            return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE

    @Gtk.Template.Callback()
    def _on_src_activated(self, _texview):
        """Called on self.src_text::active signal"""
        if not Settings.get().live_translation:
            self._on_translation()

    @Gtk.Template.Callback()
    def _on_mistakes_clicked(self, *_args):
        """Called on self.mistakes_label::activate-link signal"""
        self.mistakes.props.reveal_child = False

        translation = self.current_translation
        if translation and translation.mistakes:
            self.src_buffer.props.text = translation.mistakes.text
            # Ensure we're in the same languages
            self.src_lang_selector.selected = translation.original.src
            self.dest_lang_selector.selected = translation.original.dest

        # Run translation again
        self._on_translation()

        return Gdk.EVENT_STOP

    @Gtk.Template.Callback()
    @background_task
    async def _on_translation(self, *_args):
        if not self.provider["trans"] or self._appeared_before():
            # If it's like the last translation then it's useless to continue
            return

        # Run translation
        if self.next_translation:
            request = self.next_translation
            self.next_translation = None
        else:
            text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
            request = TranslationRequest(text, self.src_lang_selector.selected, self.dest_lang_selector.selected)

            if self.translation_loading:
                self.next_translation = request
                return

        # Show feedback for start of translation.
        self.trans_spinner.show()
        self.dest_box.props.sensitive = False
        self.langs_button_box.props.sensitive = False

        # If the two languages are the same, nothing is done
        if request.src != request.dest and request.text != "":
            self.translation_loading = True

            try:
                translation = await self.provider["trans"].translate(request)

                if translation.detected and self.src_lang_selector.selected == "auto":
                    if Settings.get().src_auto:
                        self.src_lang_selector.set_insight(
                            self.provider["trans"].normalize_lang_code(translation.detected)
                        )
                    else:
                        self.src_lang_selector.selected = translation.detected

                self.dest_buffer.props.text = translation.text

                # Finally, translation is saved in history
                self.add_history_entry(translation)

                self._check_mistakes()
                self._check_pronunciation()

            # Translation failed
            except (RequestError, ProviderError) as exc:
                self.trans_warning.props.visible = True
                self.lookup_action("copy").props.enabled = False  # type: ignore
                self.lookup_action("listen-src").props.enabled = False  # type: ignore
                self.lookup_action("listen-dest").props.enabled = False  # type: ignore

                if isinstance(exc, RequestError):
                    self.send_notification(
                        _("Translation failed, check for network issues"),
                        action={
                            "label": _("Retry"),
                            "name": "win.translation",
                        },
                    )
                elif isinstance(exc, APIKeyInvalid):
                    self.send_notification(
                        _("The provided API key is invalid"),
                        action={
                            "label": _("Retry"),
                            "name": "win.translation",
                        },
                    )
                elif isinstance(exc, APIKeyRequired):
                    self.send_notification(
                        _("API key is required to use the service"),
                        action={
                            "label": _("Preferences"),
                            "name": "app.preferences",
                        },
                    )
                else:
                    self.send_notification(
                        _("Translation failed"),
                        action={
                            "label": _("Retry"),
                            "name": "win.translation",
                        },
                    )

            else:
                self.trans_warning.props.visible = False

            finally:
                self.translation_loading = False

                if self.next_translation:
                    self._on_translation()
                else:
                    self._translation_finish()
        else:
            self.trans_mistakes = None
            self.dest_buffer.props.text = ""

            if not self.translation_loading:
                self._translation_finish()

    def _appeared_before(self):
        if not self.provider["trans"]:
            return

        src_language = self.src_lang_selector.selected
        dest_language = self.dest_lang_selector.selected
        src_text = self.src_buffer.get_text(self.src_buffer.get_start_iter(), self.src_buffer.get_end_iter(), True)
        translation = self.current_translation
        if (
            len(self.provider["trans"].history) >= self.current_history + 1
            and translation
            and (translation.original.src == src_language or "auto")
            and translation.original.dest == dest_language
            and translation.original.text == src_text
        ):
            return True
        return False

    def _translation_finish(self):
        self.trans_spinner.hide()
        self.dest_box.props.sensitive = True
        self.langs_button_box.props.sensitive = True

    """
    Provider changes functions
    """

    def _on_active_provider_changed(self, _settings: Settings, kind: str, _name: str):
        self.save_settings()
        self.reload_provider(kind)

    def _on_provider_changed(self, _settings: Gio.Settings, _key: str, name: str):
        if not self.translator_loading:
            if self.provider["trans"] and name == self.provider["trans"].name:
                self.reload_provider("translator")

            if self.provider["tts"] and name == self.provider["tts"].name:
                self.reload_provider("tts")



# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# Copyright 2023 Libretto
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import os
import typing

from gi.repository import Adw, Gio, Gtk

from dialect.define import RES_PATH
from dialect.providers import MODULES, TTS, ProviderFeature, ProvidersListModel
from dialect.settings import Settings
from dialect.widgets import ProviderPreferences

if typing.TYPE_CHECKING:
    from dialect.window import DialectWindow


@Gtk.Template(resource_path=f"{RES_PATH}/preferences.ui")
class DialectPreferencesDialog(Adw.PreferencesDialog):
    __gtype_name__ = "DialectPreferencesDialog"

    window: DialectWindow = NotImplemented

    # Child widgets
    live_translation: Adw.ExpanderRow = Gtk.Template.Child()  # type: ignore
    search_provider: Adw.SwitchRow = Gtk.Template.Child()  # type: ignore
    translate_accel: Adw.ComboRow = Gtk.Template.Child()  # type: ignore
    src_auto: Adw.SwitchRow = Gtk.Template.Child()  # type: ignore
    translator: Adw.ComboRow = Gtk.Template.Child()  # type: ignore
    translator_config: Gtk.Button = Gtk.Template.Child()  # type: ignore
    tts: Adw.ComboRow = Gtk.Template.Child()  # type: ignore
    tts_config: Gtk.Button = Gtk.Template.Child()  # type: ignore
    custom_default_font_size: Adw.ExpanderRow = Gtk.Template.Child()  # type: ignore
    default_font_size: Adw.SpinRow = Gtk.Template.Child()  # type: ignore

    def __init__(self, window: DialectWindow, **kwargs):
        super().__init__(**kwargs)

        self.window = window

        # Bind preferences with GSettings
        Settings.get().bind(
            "live-translation", self.live_translation, "enable-expansion", Gio.SettingsBindFlags.DEFAULT
        )
        Settings.get().bind("sp-translation", self.search_provider, "active", Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind("translate-accel", self.translate_accel, "selected", Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind("src-auto", self.src_auto, "active", Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind(
            "custom-default-font-size", self.custom_default_font_size, "enable-expansion", Gio.SettingsBindFlags.DEFAULT
        )
        Settings.get().bind("default-font-size", self.default_font_size, "value", Gio.SettingsBindFlags.DEFAULT)

        self.translator_config.props.sensitive = False
        self.tts_config.props.sensitive = False

        # Setup translator chooser
        trans_model = ProvidersListModel("translators")
        with self.translator.freeze_notify():
            self.translator.set_model(trans_model)
            self.translator.props.selected = trans_model.get_index_by_name(Settings.get().active_translator)
            self.translator_config.props.sensitive = self._provider_has_settings(Settings.get().active_translator)

        # Setup TTS chooser
        if len(TTS) >= 1:
            tts_model = ProvidersListModel("tts", True)
            with self.tts.freeze_notify():
                self.tts.set_model(tts_model)
                self.tts.props.selected = tts_model.get_index_by_name(Settings.get().active_tts)
                self.tts_config.props.sensitive = self._provider_has_settings(Settings.get().active_tts)
        else:
            self.tts.props.visible = False

        # Providers Settings
        self.translator_config.connect("clicked", self._open_provider, "trans")
        self.tts_config.connect("clicked", self._open_provider, "tts")

        # Translator loading
        self.window.connect("notify::translator-loading", self._on_translator_loading)

        # Search Provider
        if os.getenv("XDG_CURRENT_DESKTOP") != "GNOME":
            self.search_provider.props.visible = False

        # Connect font size signals
        self.custom_default_font_size.connect("notify::enable-expansion", self._custom_default_font_size_switch)
        self.default_font_size.get_adjustment().connect("value-changed", self._change_default_font_size)

    @Gtk.Template.Callback()
    def is_not_true(self, _widget: Gtk.Widget, boolean: bool):
        """Check if boolean is not true
        template binding closure function
        """
        return not boolean

    def _open_provider(self, _button, scope: str):
        if self.window.provider[scope] is not None:
            page = ProviderPreferences(scope, self, self.window)
            self.push_subpage(page)

    def _provider_has_settings(self, name: str):
        if not name:
            return False

        if ProviderFeature.INSTANCES in MODULES[name].features or ProviderFeature.API_KEY in MODULES[name].features:
            return True

        return False

    @Gtk.Template.Callback()
    def _switch_translator(self, _row, _value):
        """Called on self.translator::notify::selected signal"""
        provider = self.translator.get_selected_item().name  # type: ignore
        self.translator_config.props.sensitive = self._provider_has_settings(provider)
        if provider != Settings.get().active_translator:
            Settings.get().active_translator = provider

    @Gtk.Template.Callback()
    def _switch_tts(self, _row, _value):
        """Called on self.tts::notify::selected signal"""
        provider = self.tts.get_selected_item().name  # type: ignore
        self.tts_config.props.sensitive = self._provider_has_settings(provider)
        if provider != Settings.get().active_tts:
            Settings.get().active_tts = provider

    @Gtk.Template.Callback()
    def _provider_settings_tooltip(self, button: Gtk.Button, _pspec):
        if button.props.sensitive:
            button.props.tooltip_text = _("Edit Provider Settings")
        else:
            button.props.tooltip_text = _("No Settings for This Provider")

    def _on_translator_loading(self, window: DialectWindow, _value):
        self.translator.props.sensitive = not window.translator_loading
        self.tts.props.sensitive = not window.translator_loading

    def _custom_default_font_size_switch(self, row: Adw.ExpanderRow, _value):
        """Called on self.custom_default_font_size::notify::enable-expansion signal"""
        system_font_size = Settings.get().system_font_size

        if row.props.enable_expansion:
            if Settings.get().default_font_size == 0:
                # User has never set custom size before
                Settings.get().default_font_size = system_font_size
                self.default_font_size.set_value(system_font_size)
                self.window.set_font_size(system_font_size)

            else:
                self.window.set_font_size(Settings.get().default_font_size)
        else:
            self.window.set_font_size(system_font_size)
            self.custom_default_font_size.set_enable_expansion(False)

    def _change_default_font_size(self, adjustment: Gtk.Adjustment):
        """Called on self.default_font_size.get_adjustment()::value-changed signal"""
        Settings.get().default_font_size = int(adjustment.props.value)
        self.window.set_font_size(int(adjustment.props.value))



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from tempfile import NamedTemporaryFile
from urllib.parse import quote

from dialect.providers.base import (
    ProviderCapability,
    ProviderFeature,
    Translation,
    TranslationMistake,
    TranslationPronunciation,
)
from dialect.providers.errors import InvalidLangCode, UnexpectedError
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "lingva"
    prettyname = "Lingva Translate"

    capabilities = ProviderCapability.TRANSLATION | ProviderCapability.TTS
    features = (
        ProviderFeature.INSTANCES | ProviderFeature.DETECTION | ProviderFeature.MISTAKES | ProviderFeature.PRONUNCIATION
    )

    defaults = {
        "instance_url": "lingva.dialectapp.org",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 5000

    @property
    def lang_url(self):
        return self.format_url(self.instance_url, "/api/v1/languages/")

    @property
    def translate_url(self):
        return self.format_url(self.instance_url, "/api/v1/{src}/{dest}/{text}")

    @property
    def speech_url(self):
        return self.format_url(self.instance_url, "/api/v1/audio/{lang}/{text}")

    async def validate_instance(self, url):
        request = await self.get(self.format_url(url, "/api/v1/en/es/hello"), check_common=False)

        valid = False
        try:
            valid = "translation" in request
        except:  # noqa
            pass

        return valid

    async def init(self) -> None:
        response = await self.get(self.lang_url)

        try:
            if "languages" in response:
                for lang in response["languages"]:
                    if lang["code"] != "auto":
                        self.add_lang(lang["code"], lang["name"], tts=True)
            else:
                raise UnexpectedError("No langs found in server.")
        except Exception as exc:
            raise UnexpectedError from exc

    async def init_trans(self):
        await self.init()

    async def init_tts(self):
        await self.init()

    async def translate(self, request):
        src, dest = self.denormalize_lang(request.src, request.dest)
        # Format url query data
        text = quote(request.text, safe="")
        url = self.translate_url.format(text=text, src=src, dest=dest)

        # Do request
        response = await self.get(url)
        try:
            detected = response.get("info", {}).get("detectedSource", None)
            mistakes = response.get("info", {}).get("typo", None)
            src_pronunciation = response.get("info", {}).get("pronunciation", {}).get("query", None)
            dest_pronunciation = response.get("info", {}).get("pronunciation", {}).get("translation", None)

            return Translation(
                response["translation"],
                request,
                detected,
                TranslationMistake(mistakes, mistakes) if mistakes else None,
                TranslationPronunciation(src_pronunciation, dest_pronunciation),
            )

        except Exception as exc:
            raise UnexpectedError("Failed reading the translation data") from exc

    async def speech(self, text, language):
        (language,) = self.denormalize_lang(language)
        # Format url query data
        url = self.speech_url.format(text=text, lang=language)
        # Do request
        response = await self.get(url)

        try:
            file = NamedTemporaryFile()
            audio = bytearray(response["audio"])
            file.write(audio)
            file.seek(0)
            return file
        except Exception as exc:
            file.close()
            raise UnexpectedError from exc

    def check_known_errors(self, status, data):
        """Raises a proper Exception if an error is found in the data."""
        if not data:
            raise UnexpectedError("Response is empty!")

        if "error" in data:
            error = data["error"]

            if error == "Invalid target language" or error == "Invalid source language":
                raise InvalidLangCode(error)
            else:
                raise UnexpectedError(error)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gi.repository import Gio, GLib, Soup


class Session(Soup.Session):
    """
    Dialect soup session handler
    """

    instance = None
    errors = {}

    def __init__(self, *args):
        super().__init__(*args)

    @staticmethod
    def new() -> Session:
        """Create a new instance of Session."""
        s_session = Session()
        return s_session

    @staticmethod
    def get() -> Session:
        """Return an active instance of Session."""
        if Session.instance is None:
            Session.instance = Session.new()
        return Session.instance

    @staticmethod
    def get_response(session: Session, result: Gio.AsyncResult):
        try:
            response = session.send_and_read_finish(result)
            data = response.get_data()

            return data
        except GLib.Error as exc:
            raise ResponseError(exc.message) from exc


class ResponseError(Exception):
    """Exception raised when response fails."""

    def __init__(self, cause: str, message="Response has failed"):
        self.cause = cause
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.message}: {self.cause}"



# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import re
from uuid import uuid4

from dialect.providers.base import (
    ProviderCapability,
    ProviderFeature,
    ProviderLangComparison,
    Translation,
)
from dialect.providers.errors import ProviderError, UnexpectedError
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "yandex"
    prettyname = "Yandex"

    capabilities = ProviderCapability.TRANSLATION
    features = ProviderFeature.DETECTION
    lang_comp = ProviderLangComparison.DEEP

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    _headers = {
        "User-Agent": "ru.yandex.translate/21.15.4.21402814 (Xiaomi Redmi K20 Pro; Android 11)",
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 10000

        self._uuid = str(uuid4()).replace("-", "")

    @property
    def translate_url(self):
        params = {"id": self._uuid + "-0-0", "srv": "android"}
        return self.format_url("translate.yandex.net", "/api/v1/tr.json/translate", params)

    async def init_trans(self):
        # Get Yandex Translate web HTML to parse languages
        # Using `/api/v1/tr.json/getLangs` doesn't provide all the languages that Yandex supports
        html_url = self.format_url("translate.yandex.com")
        response = await self.get(html_url, check_common=False, return_json=False)

        if response:
            try:
                # Decode response bytes
                text = response.decode("utf-8")
                # Get Yandex languages
                languages = re.findall(r"TRANSLATOR_LANGS: (.*?),\n", text)[0]  # noqa
                languages: dict[str, str] = json.loads(languages)  # type: ignore
                # Get Yandex dialects list, dialects aren't valid src tranlation langs
                dialects = re.findall(r"DIALECTS: (.*?),\n", text)[0]  # noqa
                dialects: list[str] = json.loads(dialects)  # type: ignore
                # Populate languages lists
                for code, name in languages.items():
                    self.add_lang(code, name, trans_src=code not in dialects)

            except Exception as exc:
                raise UnexpectedError("Failed parsing HTML from yandex.com") from exc
        else:
            raise UnexpectedError("Could not get HTML from yandex.com")

    async def translate(self, request):
        src, dest = self.denormalize_lang(request.src, request.dest)
        # Form data
        data = {"lang": dest, "text": request.text}
        if src != "auto":
            data["lang"] = f"{src}-{dest}"

        # Do request
        response = await self.post(self.translate_url, data, self._headers, True)
        try:
            detected = None
            if "code" in response and response["code"] == 200:
                if "lang" in response:
                    detected = response["lang"].split("-")[0]
                return Translation(response["text"][0], request, detected)
            else:
                error = response["message"] if "message" in response else ""
                raise ProviderError(error)
        except Exception as exc:
            raise UnexpectedError from exc



# Copyright 2020 gi-lom
# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later


APP_ID = 'app.drey.Dialect'
PROFILE = ''
RES_PATH = '/app/drey/Dialect'
VERSION = '2.6.1'

TRANS_NUMBER = 10  # number of translations to save in history

LANG_ALIASES = {
    'iw': 'he',  # Hebrew
    'jw': 'jv',  # Javanese
    'mni-Mtei': 'mni',  # Meiteilon (Manipuri)
    'zh-CN': 'zh-Hans',
    'zh-TW': 'zh-Hant',
}

LANGUAGES = {
    "aa": "Afar",
    "ab": "Abkhazian",
    "ace": "Achinese",
    "ach": "Acoli",
    "ada": "Adangme",
    "ady": "Adyghe",
    "ae": "Avestan",
    "aeb": "Tunisian Arabic",
    "af": "Afrikaans",
    "afh": "Afrihili",
    "agq": "Aghem",
    "ain": "Ainu",
    "ak": "Akan",
    "akk": "Akkadian",
    "akz": "Alabama",
    "ale": "Aleut",
    "aln": "Gheg Albanian",
    "alt": "Southern Altai",
    "am": "Amharic",
    "an": "Aragonese",
    "ang": "Old English",
    "ann": "Obolo",
    "anp": "Angika",
    "ar": "Arabic",
    "ar-001": "Modern Standard Arabic",
    "arc": "Aramaic",
    "arn": "Mapuche",
    "aro": "Araona",
    "arp": "Arapaho",
    "arq": "Algerian Arabic",
    "ars": "Najdi Arabic",
    "ars-alt-menu": "Arabic, Najdi",
    "arw": "Arawak",
    "ary": "Moroccan Arabic",
    "arz": "Egyptian Arabic",
    "as": "Assamese",
    "asa": "Asu",
    "ase": "American Sign Language",
    "ast": "Asturian",
    "atj": "Atikamekw",
    "av": "Avaric",
    "avk": "Kotava",
    "awa": "Awadhi",
    "ay": "Aymara",
    "az": "Azerbaijani",
    "az-alt-short": "Azeri",
    "ba": "Bashkir",
    "bal": "Baluchi",
    "ban": "Balinese",
    "bar": "Bavarian",
    "bas": "Basaa",
    "bax": "Bamun",
    "bbc": "Batak Toba",
    "bbj": "Ghomala",
    "be": "Belarusian",
    "bej": "Beja",
    "bem": "Bemba",
    "bew": "Betawi",
    "bez": "Bena",
    "bfd": "Bafut",
    "bfq": "Badaga",
    "bg": "Bulgarian",
    "bgc": "Haryanvi",
    "bgn": "Western Balochi",
    "bho": "Bhojpuri",
    "bi": "Bislama",
    "bik": "Bikol",
    "bin": "Bini",
    "bjn": "Banjar",
    "bkm": "Kom",
    "bla": "Siksiká",
    "blt": "Tai Dam",
    "bm": "Bambara",
    "bn": "Bangla",
    "bo": "Tibetan",
    "bpy": "Bishnupriya",
    "bqi": "Bakhtiari",
    "br": "Breton",
    "bra": "Braj",
    "brh": "Brahui",
    "brx": "Bodo",
    "bs": "Bosnian",
    "bss": "Akoose",
    "bua": "Buriat",
    "bug": "Buginese",
    "bum": "Bulu",
    "byn": "Blin",
    "byv": "Medumba",
    "ca": "Catalan",
    "cad": "Caddo",
    "car": "Carib",
    "cay": "Cayuga",
    "cch": "Atsam",
    "ccp": "Chakma",
    "ce": "Chechen",
    "ceb": "Cebuano",
    "cgg": "Chiga",
    "ch": "Chamorro",
    "chb": "Chibcha",
    "chg": "Chagatai",
    "chk": "Chuukese",
    "chm": "Mari",
    "chn": "Chinook Jargon",
    "cho": "Choctaw",
    "chp": "Chipewyan",
    "chr": "Cherokee",
    "chy": "Cheyenne",
    "cic": "Chickasaw",
    "ckb": "Central Kurdish",
    "ckb-alt-menu": "Kurdish, Central",
    "ckb-alt-variant": "Kurdish, Sorani",
    "clc": "Chilcotin",
    "co": "Corsican",
    "cop": "Coptic",
    "cps": "Capiznon",
    "cr": "Cree",
    "crg": "Michif",
    "crh": "Crimean Tatar",
    "crj": "Southern East Cree",
    "crk": "Plains Cree",
    "crl": "Northern East Cree",
    "crm": "Moose Cree",
    "crr": "Carolina Algonquian",
    "crs": "Seselwa Creole French",
    "cs": "Czech",
    "csb": "Kashubian",
    "csw": "Swampy Cree",
    "cu": "Church Slavic",
    "cv": "Chuvash",
    "cwd": "Woods Cree",
    "cy": "Welsh",
    "da": "Danish",
    "dak": "Dakota",
    "dar": "Dargwa",
    "dav": "Taita",
    "de": "German",
    "de-AT": "Austrian German",
    "de-CH": "Swiss High German",
    "del": "Delaware",
    "den": "Slave",
    "dgr": "Dogrib",
    "din": "Dinka",
    "dje": "Zarma",
    "doi": "Dogri",
    "dsb": "Lower Sorbian",
    "dtp": "Central Dusun",
    "dua": "Duala",
    "dum": "Middle Dutch",
    "dv": "Divehi",
    "dyo": "Jola-Fonyi",
    "dyu": "Dyula",
    "dz": "Dzongkha",
    "dzg": "Dazaga",
    "ebu": "Embu",
    "ee": "Ewe",
    "efi": "Efik",
    "egl": "Emilian",
    "egy": "Ancient Egyptian",
    "eka": "Ekajuk",
    "el": "Greek",
    "elx": "Elamite",
    "en": "English",
    "en-AU": "Australian English",
    "en-CA": "Canadian English",
    "en-GB": "British English",
    "en-GB-alt-short": "UK English",
    "en-US": "American English",
    "en-US-alt-short": "US English",
    "enm": "Middle English",
    "eo": "Esperanto",
    "es": "Spanish",
    "es-419": "Latin American Spanish",
    "es-ES": "European Spanish",
    "es-MX": "Mexican Spanish",
    "esu": "Central Yupik",
    "et": "Estonian",
    "eu": "Basque",
    "ewo": "Ewondo",
    "ext": "Extremaduran",
    "fa": "Persian",
    "fa-AF": "Dari",
    "fan": "Fang",
    "fat": "Fanti",
    "ff": "Fula",
    "fi": "Finnish",
    "fil": "Filipino",
    "fit": "Tornedalen Finnish",
    "fj": "Fijian",
    "fo": "Faroese",
    "fon": "Fon",
    "fr": "French",
    "fr-CA": "Canadian French",
    "fr-CH": "Swiss French",
    "frc": "Cajun French",
    "frm": "Middle French",
    "fro": "Old French",
    "frp": "Arpitan",
    "frr": "Northern Frisian",
    "frs": "Eastern Frisian",
    "fur": "Friulian",
    "fy": "Western Frisian",
    "ga": "Irish",
    "gaa": "Ga",
    "gag": "Gagauz",
    "gan": "Gan Chinese",
    "gay": "Gayo",
    "gba": "Gbaya",
    "gbz": "Zoroastrian Dari",
    "gd": "Scottish Gaelic",
    "gez": "Geez",
    "gil": "Gilbertese",
    "gl": "Galician",
    "glk": "Gilaki",
    "gmh": "Middle High German",
    "gn": "Guarani",
    "goh": "Old High German",
    "gom": "Goan Konkani",
    "gon": "Gondi",
    "gor": "Gorontalo",
    "got": "Gothic",
    "grb": "Grebo",
    "grc": "Ancient Greek",
    "gsw": "Swiss German",
    "gu": "Gujarati",
    "guc": "Wayuu",
    "gur": "Frafra",
    "guz": "Gusii",
    "gv": "Manx",
    "gwi": "Gwichʼin",
    "ha": "Hausa",
    "hai": "Haida",
    "hak": "Hakka Chinese",
    "haw": "Hawaiian",
    "hax": "Southern Haida",
    "hdn": "Northern Haida",
    "he": "Hebrew",
    "hi": "Hindi",
    "hi-Latn": "Hindi (Latin)",
    "hi-Latn-alt-variant": "Hinglish",
    "hif": "Fiji Hindi",
    "hil": "Hiligaynon",
    "hit": "Hittite",
    "hmn": "Hmong",
    "hnj": "Hmong Njua",
    "ho": "Hiri Motu",
    "hr": "Croatian",
    "hsb": "Upper Sorbian",
    "hsn": "Xiang Chinese",
    "ht": "Haitian Creole",
    "hu": "Hungarian",
    "hup": "Hupa",
    "hur": "Halkomelem",
    "hy": "Armenian",
    "hz": "Herero",
    "ia": "Interlingua",
    "iba": "Iban",
    "ibb": "Ibibio",
    "id": "Indonesian",
    "ie": "Interlingue",
    "ig": "Igbo",
    "ii": "Sichuan Yi",
    "ik": "Inupiaq",
    "ike": "Eastern Canadian Inuktitut",
    "ikt": "Western Canadian Inuktitut",
    "ilo": "Iloko",
    "inh": "Ingush",
    "io": "Ido",
    "is": "Icelandic",
    "it": "Italian",
    "iu": "Inuktitut",
    "izh": "Ingrian",
    "ja": "Japanese",
    "jam": "Jamaican Creole English",
    "jbo": "Lojban",
    "jgo": "Ngomba",
    "jmc": "Machame",
    "jpr": "Judeo-Persian",
    "jrb": "Judeo-Arabic",
    "jut": "Jutish",
    "jv": "Javanese",
    "ka": "Georgian",
    "kaa": "Kara-Kalpak",
    "kab": "Kabyle",
    "kac": "Kachin",
    "kaj": "Jju",
    "kam": "Kamba",
    "kaw": "Kawi",
    "kbd": "Kabardian",
    "kbl": "Kanembu",
    "kcg": "Tyap",
    "kde": "Makonde",
    "kea": "Kabuverdianu",
    "ken": "Kenyang",
    "kfo": "Koro",
    "kg": "Kongo",
    "kgp": "Kaingang",
    "kha": "Khasi",
    "kho": "Khotanese",
    "khq": "Koyra Chiini",
    "khw": "Khowar",
    "ki": "Kikuyu",
    "kiu": "Kirmanjki",
    "kj": "Kuanyama",
    "kk": "Kazakh",
    "kkj": "Kako",
    "kl": "Kalaallisut",
    "kln": "Kalenjin",
    "km": "Khmer",
    "kmb": "Kimbundu",
    "kn": "Kannada",
    "ko": "Korean",
    "koi": "Komi-Permyak",
    "kok": "Konkani",
    "kos": "Kosraean",
    "kpe": "Kpelle",
    "kr": "Kanuri",
    "krc": "Karachay-Balkar",
    "kri": "Krio",
    "krj": "Kinaray-a",
    "krl": "Karelian",
    "kru": "Kurukh",
    "ks": "Kashmiri",
    "ksb": "Shambala",
    "ksf": "Bafia",
    "ksh": "Colognian",
    "ku": "Kurdish",
    "kum": "Kumyk",
    "kut": "Kutenai",
    "kv": "Komi",
    "kw": "Cornish",
    "kwk": "Kwakʼwala",
    "ky": "Kyrgyz",
    "ky-alt-variant": "Kirghiz",
    "la": "Latin",
    "lad": "Ladino",
    "lag": "Langi",
    "lah": "Western Panjabi",
    "lam": "Lamba",
    "lb": "Luxembourgish",
    "lez": "Lezghian",
    "lfn": "Lingua Franca Nova",
    "lg": "Ganda",
    "li": "Limburgish",
    "lij": "Ligurian",
    "lil": "Lillooet",
    "liv": "Livonian",
    "lkt": "Lakota",
    "lmo": "Lombard",
    "ln": "Lingala",
    "lo": "Lao",
    "lol": "Mongo",
    "lou": "Louisiana Creole",
    "loz": "Lozi",
    "lrc": "Northern Luri",
    "lsm": "Saamia",
    "lt": "Lithuanian",
    "ltg": "Latgalian",
    "lu": "Luba-Katanga",
    "lua": "Luba-Lulua",
    "lui": "Luiseno",
    "lun": "Lunda",
    "luo": "Luo",
    "lus": "Mizo",
    "luy": "Luyia",
    "lv": "Latvian",
    "lzh": "Literary Chinese",
    "lzz": "Laz",
    "mad": "Madurese",
    "maf": "Mafa",
    "mag": "Magahi",
    "mai": "Maithili",
    "mak": "Makasar",
    "man": "Mandingo",
    "mas": "Masai",
    "mde": "Maba",
    "mdf": "Moksha",
    "mdr": "Mandar",
    "men": "Mende",
    "mer": "Meru",
    "mfe": "Morisyen",
    "mg": "Malagasy",
    "mga": "Middle Irish",
    "mgh": "Makhuwa-Meetto",
    "mgo": "Metaʼ",
    "mh": "Marshallese",
    "mi": "Māori",
    "mic": "Mi'kmaq",
    "min": "Minangkabau",
    "mk": "Macedonian",
    "ml": "Malayalam",
    "mn": "Mongolian",
    "mnc": "Manchu",
    "mni": "Manipuri",
    "moe": "Innu-aimun",
    "moh": "Mohawk",
    "mos": "Mossi",
    "mr": "Marathi",
    "mrj": "Western Mari",
    "ms": "Malay",
    "mt": "Maltese",
    "mua": "Mundang",
    "mul": "Multiple languages",
    "mus": "Muscogee",
    "mwl": "Mirandese",
    "mwr": "Marwari",
    "mwv": "Mentawai",
    "my": "Burmese",
    "my-alt-variant": "Myanmar Language",
    "mye": "Myene",
    "myv": "Erzya",
    "mzn": "Mazanderani",
    "na": "Nauru",
    "nan": "Min Nan Chinese",
    "nap": "Neapolitan",
    "naq": "Nama",
    "nb": "Norwegian Bokmål",
    "nd": "North Ndebele",
    "nds": "Low German",
    "nds-NL": "Low Saxon",
    "ne": "Nepali",
    "new": "Newari",
    "ng": "Ndonga",
    "nia": "Nias",
    "niu": "Niuean",
    "njo": "Ao Naga",
    "nl": "Dutch",
    "nl-BE": "Flemish",
    "nmg": "Kwasio",
    "nn": "Norwegian Nynorsk",
    "nnh": "Ngiemboon",
    "no": "Norwegian",
    "nog": "Nogai",
    "non": "Old Norse",
    "nov": "Novial",
    "nqo": "N’Ko",
    "nr": "South Ndebele",
    "nso": "Northern Sotho",
    "nus": "Nuer",
    "nv": "Navajo",
    "nwc": "Classical Newari",
    "ny": "Nyanja",
    "nym": "Nyamwezi",
    "nyn": "Nyankole",
    "nyo": "Nyoro",
    "nzi": "Nzima",
    "oc": "Occitan",
    "oj": "Ojibwa",
    "ojb": "Northwestern Ojibwa",
    "ojc": "Central Ojibwa",
    "ojg": "Eastern Ojibwa",
    "ojs": "Oji-Cree",
    "ojw": "Western Ojibwa",
    "oka": "Okanagan",
    "om": "Oromo",
    "or": "Odia",
    "os": "Ossetic",
    "osa": "Osage",
    "ota": "Ottoman Turkish",
    "pa": "Punjabi",
    "pag": "Pangasinan",
    "pal": "Pahlavi",
    "pam": "Pampanga",
    "pap": "Papiamento",
    "pau": "Palauan",
    "pcd": "Picard",
    "pcm": "Nigerian Pidgin",
    "pdc": "Pennsylvania German",
    "pdt": "Plautdietsch",
    "peo": "Old Persian",
    "pfl": "Palatine German",
    "phn": "Phoenician",
    "pi": "Pali",
    "pis": "Pijin",
    "pl": "Polish",
    "pms": "Piedmontese",
    "pnt": "Pontic",
    "pon": "Pohnpeian",
    "pqm": "Maliseet-Passamaquoddy",
    "prg": "Prussian",
    "pro": "Old Provençal",
    "ps": "Pashto",
    "ps-alt-variant": "Pushto",
    "pt": "Portuguese",
    "pt-BR": "Brazilian Portuguese",
    "pt-PT": "European Portuguese",
    "qu": "Quechua",
    "quc": "Kʼicheʼ",
    "qug": "Chimborazo Highland Quichua",
    "raj": "Rajasthani",
    "rap": "Rapanui",
    "rar": "Rarotongan",
    "rgn": "Romagnol",
    "rhg": "Rohingya",
    "rif": "Riffian",
    "rm": "Romansh",
    "rn": "Rundi",
    "ro": "Romanian",
    "ro-MD": "Moldavian",
    "rof": "Rombo",
    "rom": "Romany",
    "rtm": "Rotuman",
    "ru": "Russian",
    "rue": "Rusyn",
    "rug": "Roviana",
    "rup": "Aromanian",
    "rw": "Kinyarwanda",
    "rwk": "Rwa",
    "sa": "Sanskrit",
    "sad": "Sandawe",
    "sah": "Yakut",
    "sam": "Samaritan Aramaic",
    "saq": "Samburu",
    "sas": "Sasak",
    "sat": "Santali",
    "saz": "Saurashtra",
    "sba": "Ngambay",
    "sbp": "Sangu",
    "sc": "Sardinian",
    "scn": "Sicilian",
    "sco": "Scots",
    "sd": "Sindhi",
    "sdc": "Sassarese Sardinian",
    "sdh": "Southern Kurdish",
    "se": "Northern Sami",
    "se-alt-menu": "Sami, Northern",
    "see": "Seneca",
    "seh": "Sena",
    "sei": "Seri",
    "sel": "Selkup",
    "ses": "Koyraboro Senni",
    "sg": "Sango",
    "sga": "Old Irish",
    "sgs": "Samogitian",
    "sh": "Serbo-Croatian",
    "shi": "Tachelhit",
    "shn": "Shan",
    "shu": "Chadian Arabic",
    "si": "Sinhala",
    "sid": "Sidamo",
    "sk": "Slovak",
    "sl": "Slovenian",
    "slh": "Southern Lushootseed",
    "sli": "Lower Silesian",
    "sly": "Selayar",
    "sm": "Samoan",
    "sma": "Southern Sami",
    "sma-alt-menu": "Sami, Southern",
    "smj": "Lule Sami",
    "smj-alt-menu": "Sami, Lule",
    "smn": "Inari Sami",
    "smn-alt-menu": "Sami, Inari",
    "sms": "Skolt Sami",
    "sms-alt-menu": "Sami, Skolt",
    "sn": "Shona",
    "snk": "Soninke",
    "so": "Somali",
    "sog": "Sogdien",
    "sq": "Albanian",
    "sr": "Serbian",
    "sr-ME": "Montenegrin",
    "srn": "Sranan Tongo",
    "srr": "Serer",
    "ss": "Swati",
    "ssy": "Saho",
    "st": "Southern Sotho",
    "stq": "Saterland Frisian",
    "str": "Straits Salish",
    "su": "Sundanese",
    "suk": "Sukuma",
    "sus": "Susu",
    "sux": "Sumerian",
    "sv": "Swedish",
    "sw": "Swahili",
    "sw-CD": "Congo Swahili",
    "swb": "Comorian",
    "syc": "Classical Syriac",
    "syr": "Syriac",
    "szl": "Silesian",
    "ta": "Tamil",
    "tce": "Southern Tutchone",
    "tcy": "Tulu",
    "te": "Telugu",
    "tem": "Timne",
    "teo": "Teso",
    "ter": "Tereno",
    "tet": "Tetum",
    "tg": "Tajik",
    "tgx": "Tagish",
    "th": "Thai",
    "tht": "Tahltan",
    "ti": "Tigrinya",
    "tig": "Tigre",
    "tiv": "Tiv",
    "tk": "Turkmen",
    "tkl": "Tokelau",
    "tkr": "Tsakhur",
    "tl": "Tagalog",
    "tlh": "Klingon",
    "tli": "Tlingit",
    "tly": "Talysh",
    "tmh": "Tamashek",
    "tn": "Tswana",
    "to": "Tongan",
    "tog": "Nyasa Tonga",
    "tok": "Toki Pona",
    "tpi": "Tok Pisin",
    "tr": "Turkish",
    "tru": "Turoyo",
    "trv": "Taroko",
    "trw": "Torwali",
    "ts": "Tsonga",
    "tsd": "Tsakonian",
    "tsi": "Tsimshian",
    "tt": "Tatar",
    "ttm": "Northern Tutchone",
    "ttt": "Muslim Tat",
    "tum": "Tumbuka",
    "tvl": "Tuvalu",
    "tw": "Twi",
    "twq": "Tasawaq",
    "ty": "Tahitian",
    "tyv": "Tuvinian",
    "tzm": "Central Atlas Tamazight",
    "udm": "Udmurt",
    "ug": "Uyghur",
    "ug-alt-variant": "Uighur",
    "uga": "Ugaritic",
    "uk": "Ukrainian",
    "umb": "Umbundu",
    "und": "Unknown language",
    "ur": "Urdu",
    "uz": "Uzbek",
    "vai": "Vai",
    "ve": "Venda",
    "vec": "Venetian",
    "vep": "Veps",
    "vi": "Vietnamese",
    "vls": "West Flemish",
    "vmf": "Main-Franconian",
    "vo": "Volapük",
    "vot": "Votic",
    "vro": "Võro",
    "vun": "Vunjo",
    "wa": "Walloon",
    "wae": "Walser",
    "wal": "Wolaytta",
    "war": "Waray",
    "was": "Washo",
    "wbp": "Warlpiri",
    "wo": "Wolof",
    "wuu": "Wu Chinese",
    "xal": "Kalmyk",
    "xh": "Xhosa",
    "xmf": "Mingrelian",
    "xog": "Soga",
    "yao": "Yao",
    "yap": "Yapese",
    "yav": "Yangben",
    "ybb": "Yemba",
    "yi": "Yiddish",
    "yo": "Yoruba",
    "yrl": "Nheengatu",
    "yue": "Cantonese",
    "yue-alt-menu": "Chinese, Cantonese",
    "za": "Zhuang",
    "zap": "Zapotec",
    "zbl": "Blissymbols",
    "zea": "Zeelandic",
    "zen": "Zenaga",
    "zgh": "Standard Moroccan Tamazight",
    "zh": "Chinese",
    "zh-alt-long": "Mandarin Chinese",
    "zh-alt-menu": "Chinese, Mandarin",
    "zh-Hans": "Simplified Chinese",
    "zh-Hans-alt-long": "Simplified Mandarin Chinese",
    "zh-Hant": "Traditional Chinese",
    "zh-Hant-alt-long": "Traditional Mandarin Chinese",
    "zu": "Zulu",
    "zun": "Zuni",
    "zxx": "No linguistic content",
    "zza": "Zaza"
}



# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from dialect.widgets.lang_selector import LangSelector  # noqa
from dialect.widgets.provider_preferences import ProviderPreferences  # noqa
from dialect.widgets.speech_button import SpeechButton  # noqa
from dialect.widgets.textview import TextView  # noqa
from dialect.widgets.theme_switcher import ThemeSwitcher  # noqa



# Copyright 2020 gi-lom
# Copyright 2020 Mufeed Ali
# Copyright 2020 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

# Initial setup
import asyncio
import logging
import sys

import gi
from gi.events import GLibEventLoopPolicy

try:
    gi.require_version("Gdk", "4.0")
    gi.require_version("Gtk", "4.0")
    gi.require_version("Gst", "1.0")
    gi.require_version("Adw", "1")
    gi.require_version("Secret", "1")
    gi.require_version("Soup", "3.0")
    gi.require_version("Spelling", "1")
    gi.require_version("GtkSource", "5")

    from gi.repository import Adw, Gio, GLib, Gst
except ImportError or ValueError:
    logging.error("Error: GObject dependencies not met.")

from dialect.define import APP_ID, RES_PATH, VERSION
from dialect.preferences import DialectPreferencesDialog
from dialect.settings import Settings
from dialect.window import DialectWindow


class Dialect(Adw.Application):
    def __init__(self):
        Adw.Application.__init__(self, application_id=APP_ID, flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)
        self.set_resource_base_path(RES_PATH)

        # App window
        self.window: DialectWindow | None = None
        # CLI
        self.argv: dict[str, str] = {}
        self._signal_handler: int | None = None

        # Add command line options
        self.add_main_option(
            "selection",
            ord("n"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            "Translate text from the primary clipboard",
            None,
        )
        self.add_main_option("text", ord("t"), GLib.OptionFlags.NONE, GLib.OptionArg.STRING, "Text to translate", None)
        self.add_main_option("src", ord("s"), GLib.OptionFlags.NONE, GLib.OptionArg.STRING, "Source lang code", None)
        self.add_main_option(
            "dest", ord("d"), GLib.OptionFlags.NONE, GLib.OptionArg.STRING, "Destination lang code", None
        )

        self.setup_actions()

    def do_activate(self):
        def on_translator_loading(_win, _pspec):
            if self.window and not self.window.translator_loading:
                # Remove signal handler
                if self._signal_handler:
                    self.window.disconnect(self._signal_handler)
                # Process CLI args
                self.process_command_line()

        self.window = self.props.active_window  # type: ignore

        if not self.window:
            width, height = Settings.get().window_size
            self.window = DialectWindow(
                application=self,
                # Translators: Do not translate the app name!
                title=_("Dialect"),
                default_height=height,
                default_width=width,
            )

        # Decide when to process command line args
        if self.window.translator_loading:
            # Wait until translator is loaded
            self._signal_handler = self.window.connect("notify::translator-loading", on_translator_loading)
        else:
            self.process_command_line()

        self.window.present()

    def do_command_line(self, command_line: Gio.ApplicationCommandLine):
        options = command_line.get_options_dict()

        # Save CLI args values
        self.argv = options.end().unpack()

        self.activate()

        return 0

    def do_startup(self):
        Adw.Application.do_startup(self)

        Gst.init(None)  # Init Gst

    def process_command_line(self):
        if not self.argv:
            return

        text = ""
        langs: dict[str, str | None] = {"src": None, "dest": None}
        selection = "selection" in self.argv

        if "text" in self.argv:
            text = self.argv["text"]
        if "src" in self.argv:
            langs["src"] = self.argv["src"]
        if "dest" in self.argv:
            langs["dest"] = self.argv["dest"]

        if self.window is not None:
            if not text and selection:
                self.window.queue_selection_translation(langs["src"], langs["dest"])
            elif text:
                self.window.translate(text, langs["src"], langs["dest"])

        # Clean CLI args
        self.argv = {}

    def setup_actions(self):
        """Setup menu actions"""

        pronunciation = Gio.SimpleAction.new_stateful("pronunciation", None, Settings.get().show_pronunciation_value)
        pronunciation.connect("change-state", self._on_pronunciation)
        self.add_action(pronunciation)

        preferences = Gio.SimpleAction(name="preferences")
        preferences.connect("activate", self._on_preferences)
        self.add_action(preferences)

        about = Gio.SimpleAction(name="about")
        about.connect("activate", self._on_about)
        self.add_action(about)

        quit_action = Gio.SimpleAction(name="quit")
        quit_action.connect("activate", self._on_quit)
        self.add_action(quit_action)

        self.set_accels_for_action("app.pronunciation", ["<Primary>P"])
        self.set_accels_for_action("app.preferences", ["<Primary>comma"])
        self.set_accels_for_action("app.quit", ["<Primary>Q"])

        self.set_accels_for_action("win.back", ["<Alt>Left"])
        self.set_accels_for_action("win.forward", ["<Alt>Right"])
        self.set_accels_for_action("win.switch", ["<Primary>S"])
        self.set_accels_for_action("win.from", ["<Primary>F"])
        self.set_accels_for_action("win.to", ["<Primary>T"])
        self.set_accels_for_action("win.clear", ["<Primary>D"])
        self.set_accels_for_action("win.font-size-inc", ["<Primary>plus", "<Primary>KP_Add"])
        self.set_accels_for_action("win.font-size-dec", ["<Primary>minus", "<Primary>KP_Subtract"])
        self.set_accels_for_action("win.paste", ["<Primary><Shift>V"])
        self.set_accels_for_action("win.copy", ["<Primary><Shift>C"])
        self.set_accels_for_action("win.listen-dest", ["<Primary>L"])
        self.set_accels_for_action("win.listen-src", ["<Primary><Shift>L"])
        self.set_accels_for_action("win.show-help-overlay", ["<Primary>question"])

    def _on_pronunciation(self, action: Gio.SimpleAction, value: GLib.Variant):
        """Update show pronunciation setting"""
        action.props.state = value
        Settings.get().show_pronunciation = value  # type: ignore

        # Update UI
        if self.window:
            self.window._check_pronunciation()

    def _on_preferences(self, _action, _param):
        """Show preferences window"""
        if self.window:
            window = DialectPreferencesDialog(self.window)
            window.present(self.window)

    def _on_about(self, _action, _param):
        """Show about dialog"""
        about = Adw.AboutDialog.new_from_appdata(f"{RES_PATH}/appdata.xml", VERSION)
        about.props.version = VERSION  # For development version
        about.props.comments = _("A translation app for GNOME.")
        about.props.copyright = _("Copyright 2020–⁠2024 The Dialect Authors")
        about.props.developers = ["Mufeed Ali", "Rafael Mardojai CM http://rafaelmardojai.com", "Libretto"]
        about.props.translator_credits = _("translator-credits")
        about.add_link(_("Donate"), "https://opencollective.com/dialect")

        about.present(self.window)

    def _on_quit(self, _action, _param):
        self.quit()


def main():
    # Set the asyncio event loop policy from PyGObject
    asyncio.set_event_loop_policy(GLibEventLoopPolicy())
    # Run the Application
    app = Dialect()
    return app.run(sys.argv)



# Copyright 2023 Mufeed Ali
# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import logging
import re
import typing

from gi.repository import Adw, GObject, Gtk

from dialect.asyncio import background_task
from dialect.define import RES_PATH
from dialect.providers import ProviderCapability, RequestError

if typing.TYPE_CHECKING:
    from dialect.window import DialectWindow


@Gtk.Template(resource_path=f"{RES_PATH}/widgets/provider_preferences.ui")
class ProviderPreferences(Adw.NavigationPage):
    __gtype_name__ = "ProviderPreferences"

    # Properties
    translation: bool = GObject.Property(type=bool, default=False)  # type: ignore
    tts: bool = GObject.Property(type=bool, default=False)  # type: ignore
    definitions: bool = GObject.Property(type=bool, default=False)  # type: ignore

    # Child widgets
    title: Adw.WindowTitle = Gtk.Template.Child()  # type: ignore
    page: Adw.PreferencesPage = Gtk.Template.Child()  # type: ignore
    instance_entry: Adw.EntryRow = Gtk.Template.Child()  # type: ignore
    instance_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    instance_reset: Gtk.Button = Gtk.Template.Child()  # type: ignore
    api_key_entry: Adw.PasswordEntryRow = Gtk.Template.Child()  # type: ignore
    api_key_stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    api_key_reset: Gtk.Button = Gtk.Template.Child()  # type: ignore
    api_usage_group: Adw.PreferencesGroup = Gtk.Template.Child()  # type: ignore
    api_usage: Gtk.LevelBar = Gtk.Template.Child()  # type: ignore
    api_usage_label: Gtk.Label = Gtk.Template.Child()  # type: ignore

    def __init__(self, scope: str, dialog: Adw.PreferencesDialog, window: DialectWindow, **kwargs):
        super().__init__(**kwargs)
        self.scope = scope
        self.provider = window.provider[scope]
        self.dialog = dialog
        self.window = window

        if self.provider:
            self.title.props.subtitle = self.provider.prettyname

            if self.provider.capabilities is not None:
                self.translation = ProviderCapability.TRANSLATION in self.provider.capabilities
                self.tts = ProviderCapability.TTS in self.provider.capabilities
                self.definitions = ProviderCapability.DEFINITIONS in self.provider.capabilities

            # Check what entries to show
            self._check_settings()

            # Load saved values
            self.instance_entry.props.text = self.provider.instance_url
            self.api_key_entry.props.text = self.provider.api_key

        # Main window progress
        self.window.connect("notify::translator-loading", self._on_translator_loading)

    def _check_settings(self):
        if not self.provider:
            return

        self.instance_entry.props.visible = self.provider.supports_instances
        self.api_key_entry.props.visible = self.provider.supports_api_key

        self.api_usage_group.props.visible = False
        if self.provider.supports_api_usage:
            self._load_api_usage()

    @background_task
    async def _load_api_usage(self):
        if not self.provider:
            return

        try:
            usage, limit = await self.provider.api_char_usage()
            level = usage / limit
            label = _("{usage:n} of {limit:n} characters").format(usage=usage, limit=limit)

            self.api_usage.props.value = level
            self.api_usage_label.props.label = label
            self.api_usage_group.props.visible = True
        except Exception as exc:
            logging.error(exc)

    @Gtk.Template.Callback()
    @background_task
    async def _on_instance_apply(self, _row):
        """Called on self.instance_entry::apply signal"""
        if not self.provider:
            return

        old_value = self.provider.instance_url
        new_value = self.instance_entry.props.text

        url = re.compile(r"https?://(www\.)?")
        self.new_instance_url = url.sub("", new_value).strip().strip("/")

        # Validate
        if self.new_instance_url != old_value:
            # Progress feedback
            self.instance_entry.props.sensitive = False
            self.api_key_entry.props.sensitive = False
            self.instance_stack.props.visible_child_name = "spinner"

            try:
                if await self.provider.validate_instance(self.new_instance_url):
                    self.provider.instance_url = self.new_instance_url
                    self.provider.reset_src_langs()
                    self.provider.reset_dest_langs()
                    self.instance_entry.remove_css_class("error")
                    self.instance_entry.props.text = self.provider.instance_url
                else:
                    self.instance_entry.add_css_class("error")
                    error_text = _("Not a valid {provider} instance")
                    error_text = error_text.format(provider=self.provider.prettyname)
                    toast = Adw.Toast(title=error_text)
                    self.dialog.add_toast(toast)
            except RequestError as exc:
                logging.error(exc)
                toast = Adw.Toast(title=_("Failed validating instance, check for network issues"))
                self.dialog.add_toast(toast)
            finally:
                self.instance_entry.props.sensitive = True
                self.api_key_entry.props.sensitive = True
                self.instance_stack.props.visible_child_name = "reset"
        else:
            self.instance_entry.remove_css_class("error")

    @Gtk.Template.Callback()
    def _on_instance_changed(self, _entry, _pspec):
        """Called on self.instance_entry::notify::text signal"""
        if not self.provider:
            return

        if self.instance_entry.props.text == self.provider.instance_url:
            self.instance_entry.props.show_apply_button = False
        elif not self.instance_entry.props.show_apply_button:
            self.instance_entry.props.show_apply_button = True

    @Gtk.Template.Callback()
    def _on_reset_instance(self, _button):
        if not self.provider:
            return

        if self.provider.instance_url != self.provider.defaults["instance_url"]:
            self.provider.reset_instance_url()

        self.instance_entry.remove_css_class("error")
        self.instance_entry.props.text = self.provider.instance_url

    @Gtk.Template.Callback()
    @background_task
    async def _on_api_key_apply(self, _row):
        """Called on self.api_key_entry::apply signal"""
        if not self.provider:
            return

        old_value = self.provider.api_key
        self.new_api_key = self.api_key_entry.get_text()

        # Validate
        if self.new_api_key != old_value:
            # Progress feedback
            self.instance_entry.props.sensitive = False
            self.api_key_entry.props.sensitive = False
            self.api_key_stack.props.visible_child_name = "spinner"

            try:
                if await self.provider.validate_api_key(self.new_api_key):
                    self.provider.api_key = self.new_api_key
                    self.api_key_entry.remove_css_class("error")
                    self.api_key_entry.props.text = self.provider.api_key
                else:
                    self.api_key_entry.add_css_class("error")
                    error_text = _("Not a valid {provider} API key")
                    error_text = error_text.format(provider=self.provider.prettyname)
                    toast = Adw.Toast(title=error_text)
                    self.dialog.add_toast(toast)
            except RequestError as exc:
                logging.error(exc)
                toast = Adw.Toast(title=_("Failed validating API key, check for network issues"))
                self.dialog.add_toast(toast)
            finally:
                self.instance_entry.props.sensitive = True
                self.api_key_entry.props.sensitive = True
                self.api_key_stack.props.visible_child_name = "reset"
        else:
            self.api_key_entry.remove_css_class("error")

    @Gtk.Template.Callback()
    def _on_reset_api_key(self, _button):
        """Called on self.api_key_reset::clicked signal"""
        if not self.provider:
            return

        if self.provider.api_key != self.provider.defaults["api_key"]:
            self.provider.reset_api_key()

        self.api_key_entry.remove_css_class("error")
        self.api_key_entry.props.text = self.provider.api_key

    def _on_translator_loading(self, window: DialectWindow, _value):
        self.page.props.sensitive = not window.translator_loading

        if not window.translator_loading:
            self.provider = self.window.provider[self.scope]
            self._check_settings()



# Copyright 2024 Mufeed Ali
# Copyright 2024 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from dialect.define import RES_PATH


@Gtk.Template(resource_path=f"{RES_PATH}/widgets/speech_button.ui")
class SpeechButton(Gtk.Button):
    __gtype_name__ = "SpeechButton"

    stack: Gtk.Stack = Gtk.Template.Child()  # type: ignore
    progress_bar: Gtk.ProgressBar = Gtk.Template.Child()  # type: ignore

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def ready(self):
        self.stack.props.visible_child_name = "ready"
        self.props.tooltip_text = _("Listen")
        self.progress_bar.props.visible = False

    def progress(self, fraction: float):
        if self.stack.props.visible_child_name != "progress":
            self.stack.props.visible_child_name = "progress"
            self.props.tooltip_text = _("Cancel Audio")
            self.progress_bar.props.visible = True

        self.progress_bar.props.fraction = fraction

    def error(self, message: str):
        self.stack.props.visible_child_name = "error"
        self.props.tooltip_text = message
        self.progress_bar.props.visible = False

    def loading(self):
        self.stack.props.visible_child_name = "loading"
        self.props.tooltip_text = _("Loading…")
        self.progress_bar.props.visible = False
